from typing import Any, Callable, Optional, Union
import uuid

from loguru import logger

from skyfi_modelship import skyfi_types as st
from .storage import download, local_folder, save_local_file, upload


def walk_fields(
    field_key: str, obj: Any, fn: Callable, **kwargs
) -> Any:
    if isinstance(obj, (list, tuple)):
        [walk_fields(f"{field_key}_{idx}", v, fn, **kwargs) for idx, v in enumerate(obj)]
    elif isinstance(obj, dict):
        for field, value in obj.items():
            walk_fields(f"{field_key}_{field}", value, fn, **kwargs)
    else:
        fn(f"{field_key}", obj, **kwargs)


def download_callable(
    request_id: Optional[uuid.UUID] = None
) -> Callable:

    def download_asset(field: str, asset: Any, **kwargs):
        if not isinstance(asset, (st.Image, st.MetadataXml)):
            return

        if not request_id:
            logger.warning("No request_id provided, skipping download: {item}", item=asset)
            return

        if not asset.path.startswith("gs://"):
            logger.warning("Image is local, skipping download: {item}", item=asset)
            return

        asset: Union[st.Image, st.MetadataXml] = asset
        folder = local_folder(request_id, field)
        path = download(asset.path, folder)
        asset.path = path

    return download_asset


def upload_callable(
    output_folder: Optional[str] = None,
    func_name: Optional[str] = None,
    request_id: Optional[uuid.UUID] = None
) -> Callable:

    def upload_asset(field, asset, **kwargs):
        if isinstance(asset, st.ImageOutput):
            asset_path = asset.value.path
        elif isinstance(asset, st.GeoJSONOutput):
            # dump the geojson to a local file
            geojson_value: st.GeoJSON = asset.value
            asset_path = save_local_file(
                request_id, field, asset.name, geojson_value.model_dump_json()
            )
            asset.path = asset_path
        else:
            # we support uploading only images and geojsons
            return

        if not output_folder:
            logger.warning("No output_folder provided, skipping upload: {item}", item=asset)
            return
        if not output_folder.startswith("gs://"):
            logger.warning("Output folder is not gs://, skipping upload: {item}", item=asset)
            return

        # upload the file to GCS
        gcs_path = upload(asset_path, output_folder,
                          func_name, name=asset.name, ref_name=asset.ref_name)

        # update the path of the objects to the GCS path
        if isinstance(asset, st.ImageOutput):
            asset.value.path = gcs_path
        elif isinstance(asset, st.GeoJSONOutput):
            asset.path = gcs_path

    return upload_asset



from loguru import logger
import uvicorn

from starlette import status

from fastapi import FastAPI, HTTPException, Request
from fastapi.exceptions import RequestValidationError
from pydantic import ValidationError

from skyfi_modelship.config import load_config

from skyfi_modelship.util.inference_request import convert_request
from skyfi_modelship.util.execution import exec_func


class FastApiHandler:

    def listen(self, func):
        logger.info("Starting fastapi for: {func}", func=func)

        config = load_config()

        fastapi = FastAPI()
        fastapi.post("/")(self.handle(func))
        uvicorn.run(fastapi, host=config.fastapi_host, port=config.fastapi_port)

    def handle(self, func):
        async def post_handler(req: Request):
            data = await req.json()
            try:
                r = convert_request(data, func)
            except ValidationError as ve:
                raise RequestValidationError(errors=ve.errors(), body=data)

            logger.info("Calling inference function for request {request_id} => {func}({kwargs})",
                        request_id=r.request_id, func=func.__name__, kwargs=r.kwargs)
            try:
                return exec_func(func, r)
            except Exception as ex:
                logger.error("Error executing function: {func}({kwargs}) => {exception}",
                             func=func.__name__, kwargs=r.kwargs, exception=ex)
                detail = f"{type(ex).__name__}: {ex}"
                raise HTTPException(status_code=status.HTTP_503_SERVICE_UNAVAILABLE, detail=detail)
        return post_handler



from builtins import int, float, str, list, dict

from .skyfi_app import SkyfiApp
from .skyfi_types import (
    Polygon,
    GeoJSON,
    Image,
    ImageType,
    MetadataXml,
    Output,
    IntOutput,
    FloatOutput,
    StrOutput,
    PolygonOutput,
    GeoJSONOutput,
    ImageOutput,
)

__version__ = "0.3.0"

__all__ = [
    SkyfiApp,
    list,
    dict,
    int,
    float,
    str,
    Polygon,
    GeoJSON,
    Image,
    ImageType,
    MetadataXml,
    Output,
    IntOutput,
    FloatOutput,
    StrOutput,
    PolygonOutput,
    GeoJSONOutput,
    ImageOutput,
]



from functools import lru_cache
from typing import Optional

from pydantic_settings import BaseSettings


class SkyfiConfig(BaseSettings):
    """
    SkyFi configuration class.
    This class contains all the configuration parameters
    for executing the model inference on SkyFi's infrastructure.
    """

    # Rabbit MQ related
    is_rabbitmq_worker: bool = False
    rabbitmq_host: Optional[str] = None
    rabbitmq_exchange: Optional[str] = None
    rabbitmq_req_queue: Optional[str] = None
    rabbitmq_resp_queue: Optional[str] = None
    rabbitmq_dl_exchange: Optional[str] = None
    rabbitmq_dl_queue: Optional[str] = None

    # FastAPI related
    is_fastapi_server: bool = False
    fastapi_host: Optional[str] = None
    fastapi_port: Optional[int] = None

    class Config:
        # Use only SKYFI_ prefixed env vars for settings
        env_prefix = "skyfi_"

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings
        ):
            """Ensures that env vars take precedence over config file."""
            return env_settings, init_settings, file_secret_settings


@lru_cache
def load_config() -> SkyfiConfig:
    return SkyfiConfig()



from enum import Enum
from functools import cached_property
import typing
from typing import Any, Dict, Generic, Optional, TypeVar
from loguru import logger
import orjson

import shapely.wkt
from pydantic import BaseModel, computed_field, field_validator
from pydantic.dataclasses import dataclass
from geojson_pydantic import FeatureCollection


@dataclass
class Polygon:
    wkt: str

    @field_validator("wkt")
    @classmethod
    def valid_wkt(cls, wkt: str):
        logger.info("Validating wkt ... {wkt}", wkt=wkt)
        try:
            shapely.wkt.loads(wkt)
        except Exception:
            raise ValueError('must be a valid wkt')
        return wkt


class GeoJSON(FeatureCollection):
    """ Store a GeoJSON Feature Object. """

    def __init__(self, geo_str: Optional[object] = None, **kwargs) -> FeatureCollection:
        if geo_str:
            logger.info("Validating geo_json ... {geo_str}", geo_str=geo_str)
            geo_json = str(geo_str)
            geo_dict = orjson.loads(geo_json)
            super().__init__(**geo_dict)
        elif len(kwargs):
            super().__init__(**kwargs)
        else:
            raise ValueError('must be a valid geo_json')


class ImageType(Enum):
    """ Supported image types. """

    GEOTIFF = "GEOTIFF"
    PNG = "PNG"


@dataclass
class Image:
    """
    Store an image path and type.

    path: will be local temporary path to the image
    type: type of the image
    """

    path: str
    type: ImageType


@dataclass
class MetadataXml:
    """
    Store an image metadata path.

    path: will be local temporary path to the metadata xml
    """

    path: str


T = TypeVar("T", int, float, str, Polygon, GeoJSON, Image, MetadataXml)


class Output(BaseModel, Generic[T]):
    """
    Base class for all outputs.

    params:
    _name_: name of the output
    _value_: the value representing the output
    _ref_name_: (Optional) reference to the input
    _tags_: (Optional) tags for the output
    _metadata_: (Optional) additional metadata
    """

    name: str
    value: T

    ref_name: Optional[str] = None
    tags: Optional[typing.List[str]] = None
    metadata: Optional[Dict[str, Any]] = None

    @computed_field
    @cached_property
    def output_type(self) -> str:
        return type(self.value).__name__


class IntOutput(Output[int]):
    pass


class FloatOutput(Output[float]):
    pass


class StrOutput(Output[str]):
    pass


class PolygonOutput(Output[Polygon]):
    pass


class GeoJSONOutput(Output[GeoJSON]):
    path: Optional[str] = None


class ImageOutput(Output[Image]):
    pass



from typing import Optional
from uuid import UUID
from google.cloud import storage
import os
from pathlib import Path

from loguru import logger


def download(path: str, folder: str) -> str:
    """ Download a file from a GCS bucket to a local folder. """

    storage_client = storage.Client()
    try:
        Path(folder).mkdir(parents=True, exist_ok=True)
        filename = os.path.basename(path)
        destination = f"{folder}/{filename}"
        blob = storage.Blob.from_string(path, client=storage_client)
        logger.info("Downloading file... {path} to {destination}",
                    path=path, destination=destination)
        blob.download_to_filename(destination)
        return destination
    except Exception as ex:
        raise ValueError(f"Error downloading file: {path}: {ex}")


def upload(path: str, folder: str,
           func_name: str, name: Optional[str], ref_name: Optional[str]) -> str:
    """ Upload a file from a local folder to a GCS bucket. """

    storage_client = storage.Client()
    try:
        dst_path = Path(path)

        # Add function name to the file name
        dst_path = dst_path.with_stem(dst_path.stem + f"_{func_name}")

        # if provided, add the name and the ref_name to the filename
        if name and ref_name:
            name_stem = f"_{name}_for_{ref_name}"
            dst_path = dst_path.with_stem(dst_path.stem + name_stem)
        dst = f"{folder}/{dst_path.name}"

        # upload the file
        logger.info("Uploading file... {path} to {dst}",
                    path=path, dst=dst)
        blob = storage.Blob.from_string(dst, client=storage_client)
        blob.upload_from_filename(path)
        return dst
    except Exception as ex:
        raise ValueError(f"Error uploading file: {path}: {ex}")


def local_folder(*args) -> str:
    # compute local folder path given the args
    str_args = [str(arg) for arg in args]
    folder = "/".join(str_args)
    return f"/tmp/{folder}"


def save_local_file(request_id: UUID, field, name: str, content: str) -> str:
    folder = local_folder(request_id, field)
    asset_path = f"{folder}/{name}.json"
    os.makedirs(folder, exist_ok=True)
    with open(asset_path, "w") as geojson_file:
        geojson_file.write(content)

    return asset_path



import shutil
from typing import Any, Optional
import uuid

from pydantic.dataclasses import dataclass

from .storage import local_folder

from .model_transformer import walk_fields, upload_callable

from .inference_request import InferenceRequest


@dataclass
class InferenceResponse:
    """
    Contains the result of the inference. request_id and output_folder are the service parameters,
    the response is the output of the inference function
    """

    request_id: uuid.UUID
    output_folder: Optional[str]
    response: Any


def exec_func(func, r: InferenceRequest) -> InferenceResponse:
    """
    Executes the inference function.
    Will post process the response, e.g. upload any output files to the output_folder.
    Will cleanup the local folder.
    """

    # call the function
    response = func(**vars(r.kwargs))
    # post process the response
    walk_fields("", response, upload_callable(r.output_folder, func.__name__, r.request_id))
    request_folder = local_folder(r.request_id)
    shutil.rmtree(request_folder, ignore_errors=True)
    return InferenceResponse(
        request_id=r.request_id, output_folder=r.output_folder, response=response
    )



from dataclasses import make_dataclass
import inspect
import sys
from typing import Optional
from loguru import logger
from simple_parsing import (
    ArgumentGenerationMode,
    ArgumentParser,
    DashVariant,
    NestedMode
)
from skyfi_modelship.util.execution import exec_func
from skyfi_modelship.util.inference_request import convert_request


class ArgsHandler:

    def handle(self, func):

        logger.info("Args Handler for: {args}", args=sys.argv)

        class ThrowingArgumentParser(ArgumentParser):
            def error(self, message):
                raise ValueError(message)

        dataclass_fields = [("request_id", str), ("output_folder", Optional[str])]
        for arg, param in inspect.signature(func).parameters.items():
            dataclass_fields.append((arg, param.annotation))

        FunctionArguments = make_dataclass('FunctionArguments', dataclass_fields)

        parser = ThrowingArgumentParser(
            description='SkyFi ModelShip Args Handler',
            exit_on_error=False,
            add_option_string_dash_variants=DashVariant.DASH,
            nested_mode=NestedMode.WITHOUT_ROOT,
            argument_generation_mode=ArgumentGenerationMode.NESTED,
        )
        parser.add_arguments(FunctionArguments, dest="func")
        try:
            args = parser.parse_args()
        except BaseException as e:
            logger.error("Error parsing arguments: {e}", e=e)
            raise e
        kwargs = args.func
        req = convert_request(vars(kwargs), func)
        logger.info(
            "Calling inference function for request {request_id} => {func}({kwargs})",
            request_id=req.request_id, func=func.__name__, kwargs=req.kwargs
        )
        return exec_func(func, req)



import inspect
from typing import Optional
import uuid

from pydantic import BaseModel, create_model

from .model_transformer import download_callable, walk_fields


class InferenceRequest(BaseModel):
    """
    Contains all information needed to run an inference function.

    request_id: internal id of the request
    output_folder: upload destination for the output files
    kwargs: the inference function parameters
    """
    request_id: uuid.UUID
    output_folder: Optional[str] = None
    kwargs: BaseModel


def convert_request(data: dict, func) -> InferenceRequest:
    """
    Converts a request dict to InferenceRequest.
    Maps all inference function parameters to skyfi.* types.
    """

    model_args = {}
    for arg, param in inspect.signature(func).parameters.items():
        model_args[arg] = (param.annotation, ...)

    FunctionArguments = create_model(
        'FunctionArguments',
        **model_args
    )
    kwargs = FunctionArguments.model_validate(data)
    walk_fields("request", vars(kwargs), download_callable(data.get("request_id")))
    request = InferenceRequest.model_validate(data | {"kwargs": kwargs})
    return request



from loguru import logger
import orjson
from fastapi.encoders import jsonable_encoder

from .config import load_config


class SkyfiApp:
    """
    The main SkyFi integration class.
    Using this class and its decorators will be all that is required
    for the integration of a model with the infrastructure.
    """

    bootstrap_func = None
    inference_func = None

    def start(self):
        """
        Start the Skyfi application.
        Execute this once all the decorators have been applied.
        """

        if self.inference_func is None:
            raise ValueError("Inference function required.")

        if self.bootstrap_func:
            logger.info("Calling the bootstrap function...")
            self.bootstrap_func()

        config = load_config()

        # if rabbitmq is enabled, import and call the handler
        if config.is_rabbitmq_worker:
            from .handler.rabbitmq_handler import RabbitMQHandler
            handler = RabbitMQHandler()
            handler.listen(self.inference_func)

        # if fastapi is enabled, import and call the handler
        elif config.is_fastapi_server:
            from .handler.fastapi_handler import FastApiHandler
            handler = FastApiHandler()
            handler.listen(self.inference_func)

        # use args
        else:
            from .handler.args_handler import ArgsHandler
            handler = ArgsHandler()
            result = handler.handle(self.inference_func)
            logger.info("Printing inference to stdout")

            print(orjson.dumps(jsonable_encoder(result)))
            return result

    def bootstrap(self, func):
        """
        Decorator to be used on a function that handles bootstrap tasks.
        These may include checking and downloading new weights, sanity checks, etc
        """

        if self.bootstrap_func is not None:
            raise ValueError("Single bootstrap function allowed.")

        self.bootstrap_func = func
        return func

    def inference(self, func):
        """
        Decorator to be used on a function that will do the model inference.
        All parameters of the functions, including the return type(s) must be
        part of the skyfi_modelship.types module. Only exception is that we allow
        iterables
        """

        if self.inference_func is not None:
            raise ValueError("Single inference function allowed.")

        self.inference_func = func
        return func



from loguru import logger
import orjson
import pika
from pika.channel import Channel
from pika.spec import Basic

from fastapi.encoders import jsonable_encoder

from skyfi_modelship.config import load_config
from skyfi_modelship.util.inference_request import convert_request
from skyfi_modelship.util.execution import exec_func


class RabbitMQHandler:

    def listen(self, func):
        config = load_config()

        connection = pika.BlockingConnection(
            pika.ConnectionParameters(config.rabbitmq_host)
        )
        channel = connection.channel()

        channel.basic_qos(prefetch_count=1)  # process 1 message at a time
        channel.basic_consume(
            queue=config.rabbitmq_req_queue, on_message_callback=self.handle(func)
        )
        logger.info("Waiting for messages...")
        channel.start_consuming()
        logger.info("Closing listener...")
        connection.close()

    def handle(self, func):
        def message_handler(
            ch: Channel,
            method: Basic.Deliver,
            properties: pika.BasicProperties,
            payload: bytes,
        ):
            logger.info(
                "Got RabbitMQ message ({properties}): {payload}",
                properties=properties,
                payload=payload,
            )
            data = orjson.loads(payload)

            try:
                r = convert_request(data, func)
            except Exception as ex:
                logger.error("Rejecting request {data}: {exc_info}", data=data, exc_info=ex)
                ch.basic_reject(delivery_tag=method.delivery_tag, requeue=False)
                return

            try:
                logger.info(
                    "Calling inference function for request {request_id} => {func}({kwargs})",
                    request_id=r.request_id,
                    func=func.__name__,
                    kwargs=r.kwargs,
                )
                response = exec_func(func, r)
                config = load_config()
                logger.info(
                    "Publishing {response} to {exchange}/{queue}",
                    response=response,
                    exchange=config.rabbitmq_exchange,
                    queue=config.rabbitmq_resp_queue,
                )
                ch.basic_publish(
                    exchange=config.rabbitmq_exchange,
                    routing_key=config.rabbitmq_resp_queue,
                    properties=pika.BasicProperties(
                        correlation_id=properties.correlation_id,
                        content_type="application/json",
                    ),
                    body=orjson.dumps(jsonable_encoder(response)),
                )
                logger.info(
                    "Acking request {request_id}: {delivery_tag}",
                    request_id=r.request_id,
                    delivery_tag=method.delivery_tag,
                )
                ch.basic_ack(delivery_tag=method.delivery_tag)
            except Exception as ex:
                logger.warning(
                    "Requeing request {request_id}, {delivery_tag}: {ex}",
                    request_id=r.request_id,
                    delivery_tag=method.delivery_tag,
                    ex=ex,
                    exc_info=ex
                )
                ch.basic_reject(delivery_tag=method.delivery_tag, requeue=True)

        return message_handler
